{-# LANGUAGE TypeOperators,
             GADTs,  
             RankNTypes,
             FlexibleContexts,
             FlexibleInstances,
             MultiParamTypeClasses,
             UndecidableInstances #-}

import Control.Monad.Trans
import Control.Monad.Identity
import Control.Monad.State
import Control.Monad.Writer
import Data.Map

type Pc m a b = Jp m a b -> Bool
type Advice m a b = (a -> m b) -> (a -> m b)
type Aspect m a b = (Pc m a b, Advice m a b)

type AEnv   m     = [Aspect m Int Int]

runAOT :: Monad m => AEnv (AOT m) -> AOT m a -> m a

law1 :: Monad m => AEnv (AOT m) -> m a -> m a
law1 aenv = runAOT aenv . lift 


class Monad m => AOPMonad m where
  wapp :: Tag -> (Int -> m Int) -> (Int -> m Int)

instance Monad m => AOPMonad (AOT m) where 
  wapp t f x = openApp t f x

i :: Tag
i = 1
j :: Tag
j = 2

f :: Int -> m Int 
g :: Int -> m Int 
f = undefined
g = undefined

prog x y = do  r1 <- wapp i f x
               r2 <- wapp j g y
               return (r1 + r2)

log' :: MonadWriter String m => Advice m a b
log' proceed x = do
  tell "Entering function ..."
  proceed x  

memo :: (MonadState (Map a b) m, Ord a) => Advice m a b
memo proceed x = do
   table <- get
   if member x table  then return (table ! x)
                  else do  y <- proceed x
                           table' <- get
                           put (insert x y table')
                           return y

test1 fPc = runAOT [(fPc, log')] (prog 5 12)

f_1 = f

c_1 f x y = do  r_1 <- f x
                r_2 <- wapp j g y
                return (r_1 + r_2)

c_ f = lift . runAOT [] . f

system' h x y = do  r_1 <- lift (h x)
                    r_2 <- wapp j g y
                    return (r_1 + r_2)

f_2 = g
c_2 h g x y = do  r_1 <- lift (h x)
                  r_2 <- g y
                  return (r_1 + r_2)


type Tag = Int
data Jp m a b = Monad m => Jp Tag

pcTag t (Jp t') = t == t'

data AOT m a = AOT { unAOT :: m (ResultAOT m a) } 

data  ResultAOT m a
      =  Return a
      |  OpenApp  Tag                    -- tag
                  (Int -> AOT m Int)     -- function
                  Int                    -- argument
                  (Int -> AOT m a)       -- continuation

instance Monad m => Monad (AOT m) where
  return   =   AOT . return . Return
  m >>= f  = AOT (unAOT m >>= \r -> case r of
     Return x        -> unAOT (f x)
     OpenApp t g x k   ->
        return (OpenApp t g x (\y -> k y >>= f)))

instance MonadTrans AOT where
  lift ma = AOT (ma >>= \ a -> (return . Return) a)    

instance MonadState s m => MonadState s (AOT m) where
  put = lift . put 
  get = lift $ get  

instance MonadWriter s m => MonadWriter s (AOT m) where
  tell = lift .tell 
  listen = undefined
  pass = undefined 

openApp t f x = AOT (return (OpenApp t f x  return))

runAOT aenv m = unAOT m >>= go where
  go (Return r)       = return r
  go (OpenApp t f x k)  =
         unAOT (  weave f aenv (Jp t) >>= 
            \woven_f -> woven_f x >>= k) >>= go

weave :: Monad m =>  (Int -> m Int) -> AEnv m ->
                     Jp m Int Int -> m (Int -> m Int)
weave f [] _                 = return f
weave f ((pc, adv):asps) jp  = 
  weave (if pc jp then (adv f) else f) asps jp

t :: Tag
t = 9

plainFib n = runAOT [(pcTag t, fibAdv)] (wapp t fibBase n)  

fibBase _ = return 1

fibAdv proceed n = 
   if (n <= 2)  then proceed n
                else do  f_1 <- wapp t fibBase (n-1)
                         f_2 <- wapp t fibBase (n-2)
                         return (f_1 + f_2)


fibAOP :: AOPMonad m => Int -> m Int
fibAOP n = case n of 
  0 -> return 1 
  1 -> return 1
  _ -> do  y <- wapp t fibAOP (n-1)
           x <- wapp t fibAOP (n-2)
           return (x+y)

plainFibAOP :: Monad m => Int -> m Int
plainFibAOP  = runAOT [] . (wapp t fibAOP)

logFibAOP :: Monad m => Int -> WriterT String m Int
logFibAOP    = runAOT [(pcTag t, log')] . (wapp t fibAOP)

memoFibAOP :: Monad m => Int -> StateT (Map Int Int) m Int
memoFibAOP   = runAOT [(pcTag t, memo)] . (wapp t fibAOP)
